#===============================================================================
# Copyright 2014-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from scipy import sparse as sp
from sklearn.utils import (check_random_state, check_array)
from sklearn.utils.sparsefuncs import mean_variance_axis
from sklearn.utils.validation import (check_is_fitted, _num_samples)
from sklearn.cluster.k_means_ import (
    k_means, _labels_inertia, _validate_center_shape)

from sklearn.utils.extmath import row_norms
import warnings

from sklearn.cluster import KMeans as KMeans_original

import daal4py
from .._utils import getFPType, get_patch_message, daal_check_version
import logging


def _tolerance(X, rtol):
    """Compute absolute tolerance from the relative tolerance"""
    if rtol == 0.0:
        return rtol
    if sp.issparse(X):
        variances = mean_variance_axis(X, axis=0)[1]
        mean_var = np.mean(variances)
    else:
        mean_var = np.var(X, axis=0).mean()
    return mean_var * rtol


def _daal4py_compute_starting_centroids(
        X, X_fptype, nClusters, cluster_centers_0, random_state):

    def is_string(s, target_str):
        return isinstance(s, str) and s == target_str

    deterministic = False
    if is_string(cluster_centers_0, 'k-means++'):
        _seed = random_state.randint(np.iinfo('i').max)
        daal_engine = daal4py.engines_mt19937(
            fptype=X_fptype, method='defaultDense', seed=_seed)
        _n_local_trials = 2 + int(np.log(nClusters))
        kmeans_init = daal4py.kmeans_init(nClusters, fptype=X_fptype,
                                          nTrials=_n_local_trials,
                                          method='plusPlusDense', engine=daal_engine)
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    elif is_string(cluster_centers_0, 'random'):
        _seed = random_state.randint(np.iinfo('i').max)
        daal_engine = daal4py.engines_mt19937(
            seed=_seed, fptype=X_fptype, method='defaultDense')
        kmeans_init = daal4py.kmeans_init(
            nClusters,
            fptype=X_fptype,
            method='randomDense',
            engine=daal_engine)
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    elif hasattr(cluster_centers_0, '__array__'):
        deterministic = True
        cc_arr = np.ascontiguousarray(cluster_centers_0, dtype=X.dtype)
        _validate_center_shape(X, nClusters, cc_arr)
        centroids_ = cc_arr
    elif callable(cluster_centers_0):
        cc_arr = cluster_centers_0(X, nClusters, random_state)
        cc_arr = np.ascontiguousarray(cc_arr, dtype=X.dtype)
        _validate_center_shape(X, nClusters, cc_arr)
        centroids_ = cc_arr
    elif is_string(cluster_centers_0, 'deterministic'):
        deterministic = True
        kmeans_init = daal4py.kmeans_init(
            nClusters, fptype=X_fptype, method='defaultDense')
        kmeans_init_res = kmeans_init.compute(X)
        centroids_ = kmeans_init_res.centroids
    else:
        raise ValueError(
            "Cluster centers should either be 'k-means++', "
            "'random', 'deterministic' or an array")
    return deterministic, centroids_


def _daal4py_kmeans_compatibility(nClusters, maxIterations, fptype="double",
                                  method="lloydDense", accuracyThreshold=0.0,
                                  resultsToEvaluate="computeCentroids"):
    kmeans_algo = None
    if daal_check_version(((2020, 'P', 2), (2021, 'B', 107))):
        kmeans_algo = daal4py.kmeans(nClusters=nClusters,
                                     maxIterations=maxIterations,
                                     fptype=fptype,
                                     resultsToEvaluate=resultsToEvaluate,
                                     accuracyThreshold=accuracyThreshold,
                                     method=method)
    else:
        assigFlag = 'computeAssignments' in resultsToEvaluate
        kmeans_algo = daal4py.kmeans(nClusters=nClusters,
                                     maxIterations=maxIterations,
                                     fptype=fptype,
                                     assignFlag=assigFlag,
                                     accuracyThreshold=accuracyThreshold,
                                     method=method)
    return kmeans_algo


def _daal4py_k_means_predict(X, nClusters, centroids,
                             resultsToEvaluate='computeAssignments'):
    X_fptype = getFPType(X)
    kmeans_algo = _daal4py_kmeans_compatibility(
        nClusters=nClusters,
        maxIterations=0,
        fptype=X_fptype,
        resultsToEvaluate=resultsToEvaluate,
        method='defaultDense')

    res = kmeans_algo.compute(X, centroids)

    return res.assignments[:, 0], res.objectiveFunction[0, 0]


def _daal4py_k_means_fit(X, nClusters, numIterations,
                         tol, cluster_centers_0, n_init, random_state):
    if numIterations < 0:
        raise ValueError("Wrong iterations number")

    X_fptype = getFPType(X)
    abs_tol = _tolerance(X, tol)  # tol is relative tolerance

    best_labels, best_inertia, best_cluster_centers = None, None, None
    best_n_iter = -1

    kmeans_algo = _daal4py_kmeans_compatibility(
        nClusters=nClusters,
        maxIterations=numIterations,
        accuracyThreshold=abs_tol,
        fptype=X_fptype,
        resultsToEvaluate='computeCentroids',
        method='defaultDense')

    for k in range(n_init):
        deterministic, starting_centroids_ = _daal4py_compute_starting_centroids(
            X, X_fptype, nClusters, cluster_centers_0, random_state)

        res = kmeans_algo.compute(X, starting_centroids_)

        inertia = res.objectiveFunction[0, 0]
        if best_inertia is None or inertia < best_inertia:
            best_cluster_centers = res.centroids
            if n_init > 1:
                best_cluster_centers = best_cluster_centers.copy()
            best_inertia = inertia

        if deterministic and n_init != 1:
            warnings.warn(
                'Explicit initial center position passed: '
                'performing only one init in k-means instead of n_init=%d'
                % n_init, RuntimeWarning, stacklevel=2)
            break

    flag_compute = 'computeAssignments|computeExactObjectiveFunction'
    best_labels, best_inertia = _daal4py_k_means_predict(
        X, nClusters, best_cluster_centers, flag_compute)
    return best_cluster_centers, best_labels, best_inertia, best_n_iter


def _fit(self, X, y=None, sample_weight=None):
    """Compute k-means clustering.

    Parameters
    ----------
    X : array-like or sparse matrix, shape=(n_samples, n_features)
        Training instances to cluster. It must be noted that the data
        will be converted to C ordering, which will cause a memory
        copy if the given data is not C-contiguous.

    y : Ignored
        not used, present here for API consistency by convention.

    sample_weight : array-like, shape (n_samples,), optional
        The weights for each observation in X. If None, all observations
        are assigned equal weight (default: None)

    """
    if self.n_init <= 0:
        raise ValueError("Invalid number of initializations."
                         " n_init=%d must be bigger than zero." % self.n_init)

    random_state = check_random_state(self.random_state)

    if self.max_iter <= 0:
        raise ValueError('Number of iterations should be a positive number,'
                         ' got %d instead' % self.max_iter)

    if self.precompute_distances == 'auto':
        precompute_distances = False
    elif isinstance(self.precompute_distances, bool):
        precompute_distances = self.precompute_distances
    else:
        raise ValueError("precompute_distances should be 'auto' or True/False"
                         ", but a value of %r was passed" %
                         self.precompute_distances)

    # avoid forcing order when copy_x=False
    order = "C" if self.copy_x else None
    X = check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32],
                    order=order, copy=self.copy_x)

    daal_ready = not sp.issparse(X) and not precompute_distances
    daal_ready = daal_ready and hasattr(X, '__array__')

    if daal_ready:
        X_len = _num_samples(X)
        daal_ready = (self.n_clusters <= X_len)
        if daal_ready and sample_weight is not None:
            sample_weight = np.asarray(sample_weight)
            daal_ready = (sample_weight.shape[0] == X_len) and (
                np.allclose(sample_weight, np.ones_like(sample_weight)))

    if not daal_ready:
        logging.info(
            "sklearn.cluster.KMeans."
            "fit: " + get_patch_message("sklearn"))
        self.cluster_centers_, self.labels_, self.inertia_, self.n_iter_ = \
            k_means(
                X, n_clusters=self.n_clusters, sample_weight=sample_weight,
                init=self.init, n_init=self.n_init, max_iter=self.max_iter,
                verbose=self.verbose, precompute_distances=precompute_distances,
                tol=self.tol, random_state=random_state, copy_x=self.copy_x,
                n_jobs=self.n_jobs, algorithm=self.algorithm,
                return_n_iter=True)
    else:
        logging.info(
            "sklearn.cluster.KMeans."
            "fit: " + get_patch_message("daal"))
        X = check_array(X, dtype=[np.float64, np.float32])
        self.cluster_centers_, self.labels_, self.inertia_, self.n_iter_ = \
            _daal4py_k_means_fit(
                X, self.n_clusters, self.max_iter, self.tol, self.init, self.n_init,
                random_state)
    return self


def _predict(self, X, sample_weight=None):
    """Predict the closest cluster each sample in X belongs to.

    In the vector quantization literature, `cluster_centers_` is called
    the code book and each value returned by `predict` is the index of
    the closest code in the code book.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape = [n_samples, n_features]
       New data to predict.

    sample_weight : array-like, shape (n_samples,), optional
        The weights for each observation in X. If None, all observations
        are assigned equal weight (default: None)

    Returns
    -------
    labels : array, shape [n_samples,]
        Index of the cluster each sample belongs to.
    """
    check_is_fitted(self, 'cluster_centers_')

    X = self._check_test_data(X)

    daal_ready = sample_weight is None and hasattr(
        X, '__array__')  # or sp.isspmatrix_csr(X)

    if daal_ready:
        logging.info(
            "sklearn.cluster.KMeans."
            "predict: " + get_patch_message("daal"))
        return _daal4py_k_means_predict(
            X, self.n_clusters, self.cluster_centers_)[0]
    logging.info(
        "sklearn.cluster.KMeans."
        "predict: " + get_patch_message("sklearn"))
    x_squared_norms = row_norms(X, squared=True)
    return _labels_inertia(X, sample_weight, x_squared_norms,
                           self.cluster_centers_)[0]


class KMeans(KMeans_original):
    __doc__ = KMeans_original.__doc__

    def __init__(self, n_clusters=8, init='k-means++', n_init=10,
                 max_iter=300, tol=1e-4, precompute_distances='auto',
                 verbose=0, random_state=None, copy_x=True,
                 n_jobs=None, algorithm='auto'):

        super(KMeans, self).__init__(
            n_clusters=n_clusters, init=init, max_iter=max_iter,
            tol=tol, precompute_distances=precompute_distances,
            n_init=n_init, verbose=verbose, random_state=random_state,
            copy_x=copy_x, n_jobs=n_jobs, algorithm=algorithm)

    def fit(self, X, y=None, sample_weight=None):
        return _fit(self, X, y=y, sample_weight=sample_weight)

    def predict(self, X, sample_weight=None):
        return _predict(self, X, sample_weight=sample_weight)



#===============================================================================
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from .validation import _daal_check_array, _daal_check_X_y


def _daal_validate_data(self, X, y=None, reset=True,
                        validate_separately=False, **check_params):
    """Validate input data and set or check the `n_features_in_` attribute.

    Parameters
    ----------
    X : {array-like, sparse matrix, dataframe} of shape \
            (n_samples, n_features)
        The input samples.
    y : array-like of shape (n_samples,), default=None
        The targets. If None, `check_array` is called on `X` and
        `check_X_y` is called otherwise.
    reset : bool, default=True
        Whether to reset the `n_features_in_` attribute.
        If False, the input will be checked for consistency with data
        provided when reset was last True.
    validate_separately : False or tuple of dicts, default=False
        Only used if y is not None.
        If False, call validate_X_y(). Else, it must be a tuple of kwargs
        to be used for calling check_array() on X and y respectively.
    **check_params : kwargs
        Parameters passed to :func:`sklearn.utils.check_array` or
        :func:`sklearn.utils.check_X_y`. Ignored if validate_separately
        is not False.

    Returns
    -------
    out : {ndarray, sparse matrix} or tuple of these
        The validated input. A tuple is returned if `y` is not None.
    """

    if y is None:
        if self._get_tags()['requires_y']:
            raise ValueError(
                f"This {self.__class__.__name__} estimator "
                f"requires y to be passed, but the target y is None."
            )
        X = _daal_check_array(X, **check_params)
        out = X
    else:
        if validate_separately:
            # We need this because some estimators validate X and y
            # separately, and in general, separately calling check_array()
            # on X and y isn't equivalent to just calling check_X_y()
            # :(
            check_X_params, check_y_params = validate_separately
            X = _daal_check_array(X, **check_X_params)
            y = _daal_check_array(y, **check_y_params)
        else:
            X, y = _daal_check_X_y(X, y, **check_params)
        out = X, y

    if check_params.get('ensure_2d', True):
        self._check_n_features(X, reset=reset)
    return out



#===============================================================================
# Copyright 2014-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import unittest
import numpy as np
import daal4py as d4p

dv = d4p._get__daal_link_version__
daal_version = tuple(map(int, (dv()[0:4], dv()[4:8])))


class Test(unittest.TestCase):
    def test_non_contig(self):
        from numpy.random import rand
        p = 10007
        nx = 1017
        ny = 77
        X = rand(p + 1, nx + 1)
        Xp = rand(p + 1, nx + 1)
        y = rand(p + 1, ny + 1)
        Xn = X[:p, :nx]
        Xpn = Xp[:p, :nx]
        yn = y[:p, :ny]
        Xc = np.ascontiguousarray(Xn)
        Xpc = np.ascontiguousarray(Xpn)
        yc = np.ascontiguousarray(yn)
        self.assertTrue(all([not Xn.flags['C_CONTIGUOUS'],
                             not Xpn.flags['C_CONTIGUOUS'],
                             not yn.flags['C_CONTIGUOUS']]))
        self.assertTrue(all([Xc.flags['C_CONTIGUOUS'],
                             Xpc.flags['C_CONTIGUOUS'],
                             yc.flags['C_CONTIGUOUS']]))
        self.assertTrue(all([np.allclose(Xc, Xn),
                             np.allclose(Xpc, Xpn),
                             np.allclose(yc, yn)]))
        regr_train = d4p.linear_regression_training()
        rtc = regr_train.compute(Xc, yc)
        regr_predict = d4p.linear_regression_prediction()
        rpc = regr_predict.compute(Xpc, rtc.model)
        regr_train = d4p.linear_regression_training()
        rtn = regr_train.compute(Xn, yn)
        regr_predict = d4p.linear_regression_prediction()
        rpn = regr_predict.compute(Xpn, rtn.model)
        self.assertTrue(np.allclose(rpn.prediction, rpc.prediction))

    def test_struct(self):
        sdata = np.array([(0.5, -1.3, 1, 100.11, 1111111),
                          (2.5, -3.3, 2, 200.22, 2222222),
                          (4.5, -5.3, 2, 350.33, 3333333),
                          (6.5, -7.3, 0, 470.44, 4444444),
                          (8.5, -9.3, 1, 270.55, 55555)],
                         dtype=[('x', 'f4'),
                                ('y', 'f4'),
                                ('categ', 'i4'),
                                ('value', 'f8'),
                                ('super', 'i8')])
        hdata = np.array([(0.5, -1.3, 1, 100.11, 1111111),
                          (2.5, -3.3, 2, 200.22, 2222222),
                          (4.5, -5.3, 2, 350.33, 3333333),
                          (6.5, -7.3, 0, 470.44, 4444444),
                          (8.5, -9.3, 1, 270.55, 55555)],
                         dtype=np.float64)
        sr = d4p.cosine_distance().compute(sdata)
        hr = d4p.cosine_distance().compute(hdata)
        self.assertTrue(np.allclose(hr.cosineDistance, sr.cosineDistance))


if __name__ == '__main__':
    unittest.main()



#===============================================================================
# Copyright 2014-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from .monkeypatch.dispatcher import enable as patch_sklearn
from .monkeypatch.dispatcher import disable as unpatch_sklearn
from .monkeypatch.dispatcher import _patch_names as sklearn_patch_names
from .monkeypatch.dispatcher import _get_map_of_algorithms as sklearn_patch_map

__all__ = [
    "patch_sklearn", "unpatch_sklearn", "sklearn_patch_names",
    "sklearn_patch_map", "cluster", "decomposition", "ensemble",
    "linear_model", "manifold", "neighbors",
    "svm", "tree", "utils", "model_selection", "metrics",
]



#===============================================================================
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py KNN scikit-learn-compatible base classes

import numpy as np
import numbers
import daal4py as d4p
from scipy import sparse as sp
from .._utils import (
    getFPType,
    sklearn_check_version,
    get_patch_message)
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import is_classifier, is_regressor
import logging
import warnings

if sklearn_check_version("0.22"):
    from sklearn.neighbors._base import KNeighborsMixin as BaseKNeighborsMixin
    from sklearn.neighbors._base import RadiusNeighborsMixin as BaseRadiusNeighborsMixin
    from sklearn.neighbors._base import NeighborsBase as BaseNeighborsBase
    from sklearn.neighbors._ball_tree import BallTree
    from sklearn.neighbors._kd_tree import KDTree
    from sklearn.neighbors._base import _check_weights
else:
    from sklearn.neighbors.base import KNeighborsMixin as BaseKNeighborsMixin
    from sklearn.neighbors.base import RadiusNeighborsMixin as BaseRadiusNeighborsMixin
    from sklearn.neighbors.base import NeighborsBase as BaseNeighborsBase
    from sklearn.neighbors.ball_tree import BallTree
    from sklearn.neighbors.kd_tree import KDTree
    from sklearn.neighbors.base import _check_weights


def training_algorithm(method, fptype, params):
    if method == 'brute':
        train_alg = d4p.bf_knn_classification_training

    else:
        train_alg = d4p.kdtree_knn_classification_training

    params['fptype'] = fptype
    return train_alg(**params)


def prediction_algorithm(method, fptype, params):
    if method == 'brute':
        predict_alg = d4p.bf_knn_classification_prediction
    else:
        predict_alg = d4p.kdtree_knn_classification_prediction

    params['fptype'] = fptype
    return predict_alg(**params)


def parse_auto_method(estimator, method, n_samples, n_features):
    result_method = method

    if (method in ['auto', 'ball_tree']):
        condition = estimator.n_neighbors is not None and \
            estimator.n_neighbors >= estimator.n_samples_fit_ // 2
        if estimator.metric == 'precomputed' or n_features > 11 or condition:
            result_method = 'brute'
        else:
            if estimator.effective_metric_ in KDTree.valid_metrics:
                result_method = 'kd_tree'
            else:
                result_method = 'brute'

    return result_method


def daal4py_fit(estimator, X, fptype):
    estimator._fit_X = X
    estimator._fit_method = estimator.algorithm
    estimator.effective_metric_ = 'euclidean'
    estimator._tree = None
    weights = getattr(estimator, 'weights', 'uniform')

    params = {
        'method': 'defaultDense',
        'k': estimator.n_neighbors,
        'voteWeights': 'voteUniform' if weights == 'uniform' else 'voteDistance',
        'resultsToCompute': 'computeIndicesOfNeighbors|computeDistances',
        'resultsToEvaluate': 'none' if getattr(estimator, '_y', None) is None
        else 'computeClassLabels'
    }
    if hasattr(estimator, 'classes_'):
        params['nClasses'] = len(estimator.classes_)

    if getattr(estimator, '_y', None) is None:
        labels = None
    else:
        labels = estimator._y.reshape(-1, 1)

    method = parse_auto_method(
        estimator, estimator.algorithm,
        estimator.n_samples_fit_, estimator.n_features_in_)
    estimator._fit_method = method
    train_alg = training_algorithm(method, fptype, params)
    estimator._daal_model = train_alg.compute(X, labels).model


def daal4py_kneighbors(estimator, X=None, n_neighbors=None,
                       return_distance=True):
    n_features = getattr(estimator, 'n_features_in_', None)
    shape = getattr(X, 'shape', None)
    if n_features and shape and len(shape) > 1 and shape[1] != n_features:
        raise ValueError((f'X has {X.shape[1]} features, '
                          f'but kneighbors is expecting {n_features} features as input'))

    if sklearn_check_version("0.22"):
        check_is_fitted(estimator)
    else:
        check_is_fitted(estimator, [])

    if n_neighbors is None:
        n_neighbors = estimator.n_neighbors
    elif n_neighbors <= 0:
        raise ValueError(
            "Expected n_neighbors > 0. Got %d" %
            n_neighbors
        )
    else:
        if not isinstance(n_neighbors, numbers.Integral):
            raise TypeError(
                "n_neighbors does not take %s value, "
                "enter integer value" %
                type(n_neighbors))

    if X is not None:
        query_is_train = False
        X = check_array(X, accept_sparse='csr', dtype=[np.float64, np.float32])
    else:
        query_is_train = True
        X = estimator._fit_X
        # Include an extra neighbor to account for the sample itself being
        # returned, which is removed later
        n_neighbors += 1

    n_samples_fit = estimator.n_samples_fit_
    if n_neighbors > n_samples_fit:
        raise ValueError(
            "Expected n_neighbors <= n_samples, "
            " but n_samples = %d, n_neighbors = %d" %
            (n_samples_fit, n_neighbors)
        )

    chunked_results = None

    try:
        fptype = getFPType(X)
    except ValueError:
        fptype = None

    weights = getattr(estimator, 'weights', 'uniform')

    params = {
        'method': 'defaultDense',
        'k': n_neighbors,
        'voteWeights': 'voteUniform' if weights == 'uniform' else 'voteDistance',
        'resultsToCompute': 'computeIndicesOfNeighbors|computeDistances',
        'resultsToEvaluate': 'none' if getattr(estimator, '_y', None) is None
        else 'computeClassLabels'
    }
    if hasattr(estimator, 'classes_'):
        params['nClasses'] = len(estimator.classes_)

    method = parse_auto_method(
        estimator, estimator._fit_method, estimator.n_samples_fit_, n_features)

    predict_alg = prediction_algorithm(method, fptype, params)
    prediction_result = predict_alg.compute(X, estimator._daal_model)

    distances = prediction_result.distances
    indices = prediction_result.indices

    if method == 'kd_tree':
        for i in range(distances.shape[0]):
            seq = distances[i].argsort()
            indices[i] = indices[i][seq]
            distances[i] = distances[i][seq]

    if return_distance:
        results = distances, indices.astype(int)
    else:
        results = indices.astype(int)

    if chunked_results is not None:
        if return_distance:
            neigh_dist, neigh_ind = zip(*chunked_results)
            results = np.vstack(neigh_dist), np.vstack(neigh_ind)
        else:
            results = np.vstack(chunked_results)

    if not query_is_train:
        return results
    # If the query data is the same as the indexed data, we would like
    # to ignore the first nearest neighbor of every sample, i.e
    # the sample itself.
    if return_distance:
        neigh_dist, neigh_ind = results
    else:
        neigh_ind = results

    n_queries, _ = X.shape
    sample_range = np.arange(n_queries)[:, None]
    sample_mask = neigh_ind != sample_range

    # Corner case: When the number of duplicates are more
    # than the number of neighbors, the first NN will not
    # be the sample, but a duplicate.
    # In that case mask the first duplicate.
    dup_gr_nbrs = np.all(sample_mask, axis=1)
    sample_mask[:, 0][dup_gr_nbrs] = False
    neigh_ind = np.reshape(
        neigh_ind[sample_mask], (n_queries, n_neighbors - 1))

    if return_distance:
        neigh_dist = np.reshape(
            neigh_dist[sample_mask], (n_queries, n_neighbors - 1))
        return neigh_dist, neigh_ind
    return neigh_ind


def validate_data(estimator, X, y=None, reset=True,
                  validate_separately=False, **check_params):
    if y is None:
        try:
            requires_y = estimator._get_tags()["requires_y"]
        except KeyError:
            requires_y = False

        if requires_y:
            raise ValueError(
                f"This {estimator.__class__.__name__} estimator "
                f"requires y to be passed, but the target y is None."
            )
        X = check_array(X, **check_params)
        out = X, y
    else:
        if validate_separately:
            # We need this because some estimators validate X and y
            # separately, and in general, separately calling check_array()
            # on X and y isn't equivalent to just calling check_X_y()
            # :(
            check_X_params, check_y_params = validate_separately
            X = check_array(X, **check_X_params)
            y = check_array(y, **check_y_params)
        else:
            X, y = check_X_y(X, y, **check_params)
        out = X, y

    if sklearn_check_version("0.23") and check_params.get('ensure_2d', True):
        estimator._check_n_features(X, reset=reset)

    return out


class NeighborsBase(BaseNeighborsBase):
    def __init__(self, n_neighbors=None, radius=None,
                 algorithm='auto', leaf_size=30, metric='minkowski',
                 p=2, metric_params=None, n_jobs=None):
        super().__init__(
            n_neighbors=n_neighbors, radius=radius,
            algorithm=algorithm, leaf_size=leaf_size, metric=metric,
            p=p, metric_params=metric_params, n_jobs=n_jobs)

    def _fit(self, X, y=None):
        if self.metric_params is not None and 'p' in self.metric_params:
            if self.p is not None:
                warnings.warn("Parameter p is found in metric_params. "
                              "The corresponding parameter from __init__ "
                              "is ignored.", SyntaxWarning, stacklevel=2)

        if hasattr(self, 'weights') and sklearn_check_version("1.0"):
            self.weights = _check_weights(self.weights)

        X_incorrect_type = isinstance(
            X, (KDTree, BallTree, NeighborsBase, BaseNeighborsBase))
        single_output = True
        self._daal_model = None
        shape = None
        correct_n_classes = True

        try:
            requires_y = self._get_tags()["requires_y"]
        except KeyError:
            requires_y = False

        if y is not None or requires_y:
            if not X_incorrect_type or y is None:
                X, y = validate_data(
                    self, X, y, accept_sparse="csr", multi_output=True,
                    dtype=[np.float64, np.float32])
                single_output = False if y.ndim > 1 and y.shape[1] > 1 else True

            shape = y.shape

            if is_classifier(self) or is_regressor(self):
                if y.ndim == 1 or y.ndim == 2 and y.shape[1] == 1:
                    self.outputs_2d_ = False
                    y = y.reshape((-1, 1))
                else:
                    self.outputs_2d_ = True

                if is_classifier(self):
                    check_classification_targets(y)
                self.classes_ = []
                self._y = np.empty(y.shape, dtype=int)
                for k in range(self._y.shape[1]):
                    classes, self._y[:, k] = np.unique(
                        y[:, k], return_inverse=True)
                    self.classes_.append(classes)

                if not self.outputs_2d_:
                    self.classes_ = self.classes_[0]
                    self._y = self._y.ravel()

                n_classes = len(self.classes_)
                if n_classes < 2:
                    correct_n_classes = False
            else:
                self._y = y
        else:
            if not X_incorrect_type:
                X, _ = validate_data(
                    self, X, accept_sparse='csr', dtype=[np.float64, np.float32])

        if not X_incorrect_type:
            self.n_samples_fit_ = X.shape[0]
            self.n_features_in_ = X.shape[1]

        try:
            fptype = getFPType(X)
        except ValueError:
            fptype = None

        weights = getattr(self, 'weights', 'uniform')

        def stock_fit(self, X, y):
            if sklearn_check_version("0.24"):
                result = super(NeighborsBase, self)._fit(X, y)
            else:
                result = super(NeighborsBase, self)._fit(X)
            return result

        if self.n_neighbors is not None:
            if self.n_neighbors <= 0:
                raise ValueError(
                    "Expected n_neighbors > 0. Got %d" %
                    self.n_neighbors
                )
            if not isinstance(self.n_neighbors, numbers.Integral):
                raise TypeError(
                    "n_neighbors does not take %s value, "
                    "enter integer value" %
                    type(self.n_neighbors))

        condition = (self.metric == 'minkowski' and self.p == 2) or \
            self.metric == 'euclidean'
        if not X_incorrect_type and weights in ['uniform', 'distance'] \
            and self.algorithm in ['brute', 'kd_tree', 'auto', 'ball_tree'] \
            and condition \
            and single_output and fptype is not None and not sp.issparse(X) and \
                correct_n_classes:
            try:
                logging.info(
                    "sklearn.neighbors.KNeighborsMixin."
                    "kneighbors: " + get_patch_message("daal"))
                daal4py_fit(self, X, fptype)
                result = self
            except RuntimeError:
                logging.info(
                    "sklearn.neighbors.KNeighborsMixin."
                    "kneighbors: " + get_patch_message("sklearn_after_daal"))
                result = stock_fit(self, X, y)
        else:
            logging.info(
                "sklearn.neighbors.KNeighborsMixin."
                "kneighbors: " + get_patch_message("sklearn"))
            result = stock_fit(self, X, y)

        if y is not None and is_regressor(self):
            self._y = y if shape is None else y.reshape(shape)

        return result


class KNeighborsMixin(BaseKNeighborsMixin):
    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        daal_model = getattr(self, '_daal_model', None)
        if X is not None:
            X = check_array(
                X, accept_sparse='csr', dtype=[
                    np.float64, np.float32])
        x = self._fit_X if X is None else X
        try:
            fptype = getFPType(x)
        except ValueError:
            fptype = None

        if daal_model is not None and fptype is not None and not sp.issparse(
                X):
            logging.info(
                "sklearn.neighbors.KNeighborsMixin."
                "kneighbors: " + get_patch_message("daal"))
            result = daal4py_kneighbors(self, X, n_neighbors, return_distance)
        else:
            logging.info(
                "sklearn.neighbors.KNeighborsMixin."
                "kneighbors:" + get_patch_message("sklearn"))
            if daal_model is not None or getattr(self, '_tree', 0) is None and \
                    self._fit_method == 'kd_tree':
                if sklearn_check_version("0.24"):
                    BaseNeighborsBase._fit(self, self._fit_X, getattr(self, '_y', None))
                else:
                    BaseNeighborsBase._fit(self, self._fit_X)
            result = super(KNeighborsMixin, self).kneighbors(
                X, n_neighbors, return_distance)

        return result


class RadiusNeighborsMixin(BaseRadiusNeighborsMixin):
    def radius_neighbors(self, X=None, radius=None, return_distance=True,
                         sort_results=False):
        daal_model = getattr(self, '_daal_model', None)

        if daal_model is not None or getattr(self, '_tree', 0) is None and \
                self._fit_method == 'kd_tree':
            if sklearn_check_version("0.24"):
                BaseNeighborsBase._fit(self, self._fit_X, getattr(self, '_y', None))
            else:
                BaseNeighborsBase._fit(self, self._fit_X)
        if sklearn_check_version("0.22"):
            result = BaseRadiusNeighborsMixin.radius_neighbors(
                self, X, radius, return_distance, sort_results)
        else:
            result = BaseRadiusNeighborsMixin.radius_neighbors(
                self, X, radius, return_distance)

        return result



#===============================================================================
# Copyright 2014-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py PCA example for distributed memory systems; SPMD mode
# run like this:
#    mpirun -n 4 python ./pca_spmd.py

import daal4py as d4p
from numpy import loadtxt, allclose

if __name__ == "__main__":
    # Initialize SPMD mode
    d4p.daalinit()

    # Each process gets its own data
    infile = "./data/distributed/pca_normalized_" + str(d4p.my_procid() + 1) + ".csv"

    # configure a PCA object to use svd instead of default correlation
    algo = d4p.pca(method='svdDense', distributed=True)
    # let's provide a file directly, not a table/array
    result1 = algo.compute(infile)

    # We can also load the data ourselfs and provide the numpy array
    data = loadtxt(infile, delimiter=',')
    result2 = algo.compute(data)

    # PCA result objects provide eigenvalues, eigenvectors, means and variances
    assert allclose(result1.eigenvalues, result2.eigenvalues)
    assert allclose(result1.eigenvectors, result2.eigenvectors)
    assert result1.means is None and \
           result2.means is None or \
           allclose(result1.means, result2.means)
    assert result1.variances is None and \
           result2.variances is None or \
           allclose(result1.variances, result2.variances)

    print('All looks good!')
    d4p.daalfini()



#===============================================================================
# Copyright 2014-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

# daal4py Decision Forest Classification example for shared memory systems

import daal4py as d4p
import numpy as np

# let's try to use pandas' fast csv reader
try:
    import pandas

    def read_csv(f, c, t=np.float64):
        return pandas.read_csv(f, usecols=c, delimiter=',', header=None, dtype=t)
except ImportError:
    # fall back to numpy loadtxt
    def read_csv(f, c, t=np.float64):
        return np.loadtxt(f, usecols=c, delimiter=',', ndmin=2, dtype=t)


def main(readcsv=read_csv, method='defaultDense'):
    # input data file
    infile = "./data/batch/df_classification_train.csv"
    testfile = "./data/batch/df_classification_test.csv"

    # Configure a training object (5 classes)
    train_algo = d4p.decision_forest_classification_training(
        5,
        method=method,
        nTrees=10,
        minObservationsInLeafNode=8,
        featuresPerNode=3,
        engine=d4p.engines_mt19937(seed=777),
        varImportance='MDI',
        bootstrap=True,
        resultsToCompute='computeOutOfBagError'
    )

    # Read data. Let's use 3 features per observation
    data = readcsv(infile, range(3), t=np.float32)
    labels = readcsv(infile, range(3, 4), t=np.float32)
    train_result = train_algo.compute(data, labels)
    # Traiing result provides (depending on parameters) model,
    # outOfBagError, outOfBagErrorPerObservation and/or variableImportance

    # Now let's do some prediction
    predict_algo = d4p.decision_forest_classification_prediction(
        nClasses=5,
        resultsToEvaluate="computeClassLabels|computeClassProbabilities",
        votingMethod="unweighted"
    )
    # read test data (with same #features)
    pdata = readcsv(testfile, range(3), t=np.float32)
    plabels = readcsv(testfile, range(3, 4), t=np.float32)
    # now predict using the model from the training above
    predict_result = predict_algo.compute(pdata, train_result.model)

    # Prediction result provides prediction
    assert(predict_result.prediction.shape == (pdata.shape[0], 1))

    return (train_result, predict_result, plabels)


if __name__ == "__main__":
    (train_result, predict_result, plabels) = main()
    print("\nVariable importance results:\n", train_result.variableImportance)
    print("\nOOB error:\n", train_result.outOfBagError)
    print(
        "\nDecision forest prediction results (first 10 rows):\n",
        predict_result.prediction[0:10]
    )
    print(
        "\nDecision forest probabilities results (first 10 rows):\n",
        predict_result.probabilities[0:10]
    )
    print("\nGround truth (first 10 rows):\n", plabels[0:10])
    print('All looks good!')



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from cpython.ref cimport PyObject

include "kernel_functions.pxi"

cdef class PyLinearKernelParams:
    cdef linear_kernel_params pt

    def __init__(self, scale, shift):
        self.pt.scale = scale
        self.pt.shift = shift

    @property
    def scale(self):
        return self.pt.scale

    @scale.setter
    def scale(self,val):
        self.pt.scale = val

    @property
    def shift(self):
        return self.pt.shift

    @shift.setter
    def shift(self,val):
        self.pt.shift = val


cdef class PyLinearKernelCompute:
    cdef linear_kernel_compute * thisptr

    def __cinit__(self, PyLinearKernelParams params):
        self.thisptr = new linear_kernel_compute(&params.pt)

    def __dealloc__(self):
        del self.thisptr

    def compute(self, x, y):
        self.thisptr.compute(<PyObject *>x, <PyObject *>y)

    def get_values(self):
        return <object>self.thisptr.get_values()

cdef class PyRbfKernelParams:
    cdef rbf_kernel_params pt

    def __init__(self, sigma):
        self.pt.sigma = sigma

    @property
    def sigma(self):
        return self.pt.sigma

    @sigma.setter
    def sigma(self,val):
        self.pt.sigma = val


cdef class PyRbfKernelCompute:
    cdef rbf_kernel_compute * thisptr

    def __cinit__(self, PyRbfKernelParams params):
        self.thisptr = new rbf_kernel_compute(&params.pt)

    def __dealloc__(self):
        del self.thisptr

    def compute(self, x, y):
        self.thisptr.compute(<PyObject *>x, <PyObject *>y)

    def get_values(self):
        return <object>self.thisptr.get_values()

cdef class PyPolyKernelParams:
    cdef polynomial_kernel_params pt

    def __init__(self, scale, shift, degree):
        self.pt.scale = scale
        self.pt.shift = shift
        self.pt.degree = degree


cdef class PyPolyKernelCompute:
    cdef polynomial_kernel_compute * thisptr

    def __cinit__(self, PyPolyKernelParams params):
        self.thisptr = new polynomial_kernel_compute(&params.pt)

    def __dealloc__(self):
        del self.thisptr

    def compute(self, x, y):
        self.thisptr.compute(<PyObject *>x, <PyObject *>y)

    def get_values(self):
        return <object>self.thisptr.get_values()



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import cython
from libcpp cimport bool
from cpython.ref cimport PyObject
cimport numpy as npc


include "svm.pxi"

cdef class PySvmParams:
    cdef svm_params pt

    def __init__(self, method, kernel, class_count, c, nu,
                 epsilon, accuracy_threshold, max_iteration_count,
                 cache_size, shrinking, tau, sigma, shift, scale, degree):
        self.pt.method = to_std_string( < PyObject * >method)
        self.pt.kernel = to_std_string( < PyObject * >kernel)
        self.pt.class_count = class_count
        self.pt.c = c
        self.pt.nu = nu
        self.pt.epsilon = epsilon
        self.pt.accuracy_threshold = accuracy_threshold
        self.pt.max_iteration_count = max_iteration_count
        self.pt.cache_size = cache_size
        self.pt.shrinking = shrinking
        self.pt.tau = tau
        self.pt.sigma = sigma
        self.pt.shift = shift
        self.pt.scale = scale
        self.pt.degree = degree


cdef class PyClassificationSvmModel:
    cdef svm_model[classification] * thisptr

    def __cinit__(self):
        self.thisptr = new svm_model[classification]()

    def __dealloc__(self):
        del self.thisptr

    def __setstate__(self, state):
        if isinstance(state, bytes):
           self.thisptr.deserialize(state)
        else:
           raise ValueError("Invalid state")

    def __getstate__(self):
        if self.thisptr == NULL:
            raise ValueError("Pointer to svm model is NULL")
        bytes = self.thisptr.serialize()
        return bytes


cdef class PyClassificationSvmTrain:
    cdef svm_train[classification] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_train[classification](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def train(self, data, labels, weights):
        self.thisptr.train( < PyObject * >data, < PyObject * >labels, < PyObject * >weights)

    def get_support_vectors(self):
        return < object > self.thisptr.get_support_vectors()

    def get_support_indices(self):
        return < object > self.thisptr.get_support_indices()

    def get_coeffs(self):
        return < object > self.thisptr.get_coeffs()

    def get_biases(self):
        return < object > self.thisptr.get_biases()

    def get_model(self):
        cdef PyClassificationSvmModel res = PyClassificationSvmModel.__new__(PyClassificationSvmModel)
        res.thisptr[0] = self.thisptr.get_model()
        return res


cdef class PyClassificationSvmInfer:
    cdef svm_infer[classification] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_infer[classification](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def infer(self, data, PyClassificationSvmModel model):
        self.thisptr.infer( < PyObject * >data, model.thisptr)

    def infer_builder(self, data, support_vectors, coeffs, biases):
        self.thisptr.infer( < PyObject * >data, < PyObject * >support_vectors, < PyObject * >coeffs, < PyObject * >biases)

    def get_labels(self):
        return < object > self.thisptr.get_labels()

    def get_decision_function(self):
        return < object > self.thisptr.get_decision_function()


cdef class PyRegressionSvmModel:
    cdef svm_model[regression] * thisptr

    def __cinit__(self):
        self.thisptr = new svm_model[regression]()

    def __dealloc__(self):
        del self.thisptr

    def __setstate__(self, state):
        if isinstance(state, bytes):
           self.thisptr.deserialize(state)
        else:
           raise ValueError("Invalid state")

    def __getstate__(self):
        if self.thisptr == NULL:
            raise ValueError("Pointer to svm model is NULL")
        bytes = self.thisptr.serialize()
        return bytes


cdef class PyRegressionSvmTrain:
    cdef svm_train[regression] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_train[regression](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def train(self, data, labels, weights):
        self.thisptr.train( < PyObject * >data, < PyObject * >labels, < PyObject * >weights)

    def get_support_vectors(self):
        return < object > self.thisptr.get_support_vectors()

    def get_support_indices(self):
        return < object > self.thisptr.get_support_indices()

    def get_coeffs(self):
        return < object > self.thisptr.get_coeffs()

    def get_biases(self):
        return < object > self.thisptr.get_biases()

    def get_model(self):
        cdef PyRegressionSvmModel res = PyRegressionSvmModel.__new__(PyRegressionSvmModel)
        res.thisptr[0] = self.thisptr.get_model()
        return res


cdef class PyRegressionSvmInfer:
    cdef svm_infer[regression] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_infer[regression](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def infer(self, data, PyRegressionSvmModel model):
        self.thisptr.infer( < PyObject * >data, model.thisptr)

    def infer_builder(self, data, support_vectors, coeffs, biases):
        self.thisptr.infer( < PyObject * >data, < PyObject * >support_vectors, < PyObject * >coeffs, < PyObject * >biases)

    def get_labels(self):
        return < object > self.thisptr.get_labels()


cdef class PyNuClassificationSvmModel:
    cdef svm_model[nu_classification] * thisptr

    def __cinit__(self):
        self.thisptr = new svm_model[nu_classification]()

    def __dealloc__(self):
        del self.thisptr

    def __setstate__(self, state):
        if isinstance(state, bytes):
           self.thisptr.deserialize(state)
        else:
           raise ValueError("Invalid state")

    def __getstate__(self):
        if self.thisptr == NULL:
            raise ValueError("Pointer to svm model is NULL")
        bytes = self.thisptr.serialize()
        return bytes


cdef class PyNuClassificationSvmTrain:
    cdef svm_train[nu_classification] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_train[nu_classification](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def train(self, data, labels, weights):
        self.thisptr.train( < PyObject * >data, < PyObject * >labels, < PyObject * >weights)

    def get_support_vectors(self):
        return < object > self.thisptr.get_support_vectors()

    def get_support_indices(self):
        return < object > self.thisptr.get_support_indices()

    def get_coeffs(self):
        return < object > self.thisptr.get_coeffs()

    def get_biases(self):
        return < object > self.thisptr.get_biases()

    def get_model(self):
        cdef PyNuClassificationSvmModel res = PyNuClassificationSvmModel.__new__(PyNuClassificationSvmModel)
        res.thisptr[0] = self.thisptr.get_model()
        return res


cdef class PyNuClassificationSvmInfer:
    cdef svm_infer[nu_classification] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_infer[nu_classification](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def infer(self, data, PyNuClassificationSvmModel model):
        self.thisptr.infer( < PyObject * >data, model.thisptr)

    def infer_builder(self, data, support_vectors, coeffs, biases):
        self.thisptr.infer( < PyObject * >data, < PyObject * >support_vectors, < PyObject * >coeffs, < PyObject * >biases)

    def get_labels(self):
        return < object > self.thisptr.get_labels()

    def get_decision_function(self):
        return < object > self.thisptr.get_decision_function()


cdef class PyNuRegressionSvmModel:
    cdef svm_model[nu_regression] * thisptr

    def __cinit__(self):
        self.thisptr = new svm_model[nu_regression]()

    def __dealloc__(self):
        del self.thisptr

    def __setstate__(self, state):
        if isinstance(state, bytes):
           self.thisptr.deserialize(state)
        else:
           raise ValueError("Invalid state")

    def __getstate__(self):
        if self.thisptr == NULL:
            raise ValueError("Pointer to svm model is NULL")
        bytes = self.thisptr.serialize()
        return bytes


cdef class PyNuRegressionSvmTrain:
    cdef svm_train[nu_regression] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_train[nu_regression](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def train(self, data, labels, weights):
        self.thisptr.train( < PyObject * >data, < PyObject * >labels, < PyObject * >weights)

    def get_support_vectors(self):
        return < object > self.thisptr.get_support_vectors()

    def get_support_indices(self):
        return < object > self.thisptr.get_support_indices()

    def get_coeffs(self):
        return < object > self.thisptr.get_coeffs()

    def get_biases(self):
        return < object > self.thisptr.get_biases()

    def get_model(self):
        cdef PyNuRegressionSvmModel res = PyNuRegressionSvmModel.__new__(PyNuRegressionSvmModel)
        res.thisptr[0] = self.thisptr.get_model()
        return res


cdef class PyNuRegressionSvmInfer:
    cdef svm_infer[nu_regression] * thisptr

    def __cinit__(self, PySvmParams params):
        self.thisptr = new svm_infer[nu_regression](& params.pt)

    def __dealloc__(self):
        del self.thisptr

    def infer(self, data, PyNuRegressionSvmModel model):
        self.thisptr.infer( < PyObject * >data, model.thisptr)

    def infer_builder(self, data, support_vectors, coeffs, biases):
        self.thisptr.infer( < PyObject * >data, < PyObject * >support_vectors, < PyObject * >coeffs, < PyObject * >biases)

    def get_labels(self):
        return < object > self.thisptr.get_labels()



#===============================================================================
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import numpy as np
from onedal.common import _check_array

try:
    from _onedal4py_dpc import (
        PyLinearKernelParams,
        PyLinearKernelCompute,
        PyRbfKernelParams,
        PyRbfKernelCompute,
        PyPolyKernelParams,
        PyPolyKernelCompute,
    )
except ImportError:
    from _onedal4py_host import (
        PyLinearKernelParams,
        PyLinearKernelCompute,
        PyRbfKernelParams,
        PyRbfKernelCompute,
        PyPolyKernelParams,
        PyPolyKernelCompute,
    )


def linear_kernel(X, Y=None, scale=1.0, shift=0.0):
    """
    Compute the linear kernel between X and Y:

        K(x, y) = scale*dot(x, y^T) + shift

    for each pair of rows x in X and y in Y.

    Parameters
    ----------
    X : ndarray of shape (n_samples_X, n_features)

    Y : ndarray of shape (n_samples_Y, n_features)

    scale : float, default=1.0

    shift : float, default=0.0

    Returns
    -------
    kernel_matrix : ndarray of shape (n_samples_X, n_samples_Y)
    """

    X = _check_array(X, dtype=[np.float64, np.float32], force_all_finite=False)
    if Y is None:
        Y = X
    else:
        Y = _check_array(
            Y, dtype=[np.float64, np.float32], force_all_finite=False)

    _onedal_params = PyLinearKernelParams(scale, shift)
    c_kernel = PyLinearKernelCompute(_onedal_params)
    c_kernel.compute(X, Y)
    return c_kernel.get_values()


def rbf_kernel(X, Y=None, gamma=None):
    """
    Compute the rbf (gaussian) kernel between X and Y:

        K(x, y) = exp(-gamma ||x-y||^2)

    for each pair of rows x in X and y in Y.

    Parameters
    ----------
    X : ndarray of shape (n_samples_X, n_features)

    Y : ndarray of shape (n_samples_Y, n_features)

    gamma : float, default=None
        If None, defaults to 1.0 / n_features.

    Returns
    -------
    kernel_matrix : ndarray of shape (n_samples_X, n_samples_Y)
    """

    X = _check_array(X, dtype=[np.float64, np.float32], force_all_finite=False)
    if Y is None:
        Y = X
    else:
        Y = _check_array(
            Y, dtype=[np.float64, np.float32], force_all_finite=False)

    if gamma is None:
        gamma = 1.0 / X.shape[1]

    sigma = np.sqrt(0.5 / gamma)
    _onedal_params = PyRbfKernelParams(sigma=sigma)
    c_kernel = PyRbfKernelCompute(_onedal_params)
    c_kernel.compute(X, Y)
    return c_kernel.get_values()


def poly_kernel(X, Y=None, gamma=1.0, coef0=0.0, degree=3):
    """
    Compute the poly kernel between X and Y:

        K(x, y) = (scale*dot(x, y^T) + shift)**degree

    for each pair of rows x in X and y in Y.

    Parameters
    ----------
    X : ndarray of shape (n_samples_X, n_features)

    Y : ndarray of shape (n_samples_Y, n_features)

    scale : float, default=1.0

    shift : float, default=0.0

    degree : float, default=3

    Returns
    -------
    kernel_matrix : ndarray of shape (n_samples_X, n_samples_Y)
    """

    X = _check_array(X, dtype=[np.float64, np.float32], force_all_finite=False)
    if Y is None:
        Y = X
    else:
        Y = _check_array(
            Y, dtype=[np.float64, np.float32], force_all_finite=False)

    _onedal_params = PyPolyKernelParams(
        scale=gamma, shift=coef0, degree=degree)
    c_kernel = PyPolyKernelCompute(_onedal_params)
    c_kernel.compute(X, Y)
    return c_kernel.get_values()
